
import logging

import socket

from . import connection


def wait_for_client_connection(port):

    sock = socket.socket(family=socket.AF_INET6)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(('', port))
    sock.listen(0)
    conn, address = sock.accept()

    logging.info(f"Accepted connection from client: {address}")

    return conn


def listen(port, handler, *args, **kwargs):
    """
    Listens on a given port and invokes the handler function with a connection and the provided args and kwargs
    :param port: Port on which to listen
    :param handler: Callable which takes a connection and the remaining args
    :param args:
    :param kwargs:
    """
    logging.debug(f"Starting external Python module '{handler.__name__}' in state: [PASSIVE]")
    logging.debug(f"Waiting for connection from client on port: {port}")

    with connection.Connection(wait_for_client_connection(port)) as conn:
        handler(conn, *args, **kwargs)




import ctypes
import logging

import numpy as np

from ismrmrd import Acquisition, Waveform

from ..external.constants import uint64
from gadgetron.external.readers import read, read_acquisition_header, read_vector, read_waveform_header
from gadgetron.external.writers import write_optional, write_array, write_object_array, write_acquisition_header


class AcquisitionBucketStats:

    def __init__(self, kspace_encode_step_1={}, kspace_encode_step_2={}, slice={}, phase={}, contrast={}, repetition={},
                 set={}, segment={}, average={}):
        self.kspace_encode_step_1 = kspace_encode_step_1
        self.kspace_encode_step_2 = kspace_encode_step_2
        self.contrast = contrast
        self.slice = slice
        self.phase = phase
        self.repetition = repetition
        self.segment = segment
        self.average = average
        self.set = set


class AcquisitionBucket:
    def __init__(self, data, datastats, ref, refstats, waveforms):
        self.data = data
        self.datastats = datastats
        self.ref = ref
        self.refstats = refstats
        self.waveforms = waveforms


class bundle_meta(ctypes.Structure):
    _fields_ = [
        ("count", ctypes.c_uint64),
        ("header_bytes", ctypes.c_uint64),
        ("data_bytes", ctypes.c_uint64),
        ("trajectory_bytes", ctypes.c_uint64)

    ]


class stats_meta(ctypes.Structure):
    _fields_ = [
        ("nbytes", ctypes.c_uint64)
    ]


class waveform_meta(ctypes.Structure):
    _fields_ = [
        ("count", ctypes.c_uint64),
        ("header_bytes", ctypes.c_uint64),
        ("data_bytes", ctypes.c_uint64)
    ]


class bucket_meta(ctypes.Structure):
    _fields_ = [
        ("data", bundle_meta),
        ("reference", bundle_meta),
        ("data_stats", stats_meta),
        ("reference_stats", stats_meta),
        ("waveforms", waveform_meta)
    ]


def read_bucketstats(source):
    count = read(source, uint64)
    return [AcquisitionBucketStats(*[{s for s in read_vector(source, np.uint16)}
                                     for _ in range(9)])
            for _ in range(count)]


def read_waveforms(source, sizes):
    headers = [read_waveform_header(source) for _ in range(sizes.count)]
    data_arrays = [read_data_as_array(source, np.uint32, (header.channels, header.number_of_samples))
                   for header in headers]
    return [Waveform(head, data) for head, data in zip(headers, data_arrays)]


def read_data_as_array(source, data_type, shape):
    dtype = np.dtype(data_type)
    bytesize = np.prod(shape) * dtype.itemsize
    return np.reshape(np.frombuffer(source.read(bytesize), dtype), shape)


def read_acquisitions(source, sizes):
    headers = [read_acquisition_header(source) for _ in range(sizes.count)]

    trajectories = [read_data_as_array(source, np.float32, (head.number_of_samples, head.trajectory_dimensions))
                    if head.trajectory_dimensions > 0 else None
                    for head in headers]

    acqs = [read_data_as_array(source, np.complex64, (head.active_channels, head.number_of_samples))
            for head in headers]

    return [Acquisition(header, data, trajectory) for header, data, trajectory in zip(headers, acqs, trajectories)]


def read_acquisition_bucket(source):
    meta = bucket_meta.from_buffer_copy(source.read(ctypes.sizeof(bucket_meta)))

    return AcquisitionBucket(
        read_acquisitions(source, meta.data),
        read_bucketstats(source),
        read_acquisitions(source, meta.reference),
        read_bucketstats(source),
        read_waveforms(source, meta.waveforms)
    )






import numpy

from ..external import constants


def write_optional(destination, optional, continuation, *args, **kwargs):
    if optional is None:
        destination.write(constants.bool.pack(False))
    else:
        destination.write(constants.bool.pack(True))
        continuation(destination, optional, *args, **kwargs)


def write_vector(destination, values, type=constants.uint64):
    destination.write(constants.uint64.pack(len(values)))
    for val in values:
        destination.write(type.pack(val))


def write_array(destination, array):
    write_vector(destination, array.shape)
    destination.write(array.tobytes(order='F'))


def write_object_array(destination, array, writer, *args, **kwargs):
    write_vector(destination, array.shape)
    for item in numpy.nditer(array, ('refs_ok', 'zerosize_ok'), order='F'):
        item = item.item()  # Get rid of the numpy 0-dimensional array.
        writer(destination, item, *args, **kwargs)


def write_acquisition_header(destination, header):
    destination.write(header)


def write_image_header(destination, header):
    destination.write(header)


def write_byte_string(destination, byte_string, type=constants.uint32):
    destination.write(type.pack(len(byte_string)))
    destination.wtite(byte_string)


def write_acquisition(destination, acquisition):
    message_id_bytes = constants.GadgetMessageIdentifier.pack(constants.GADGET_MESSAGE_ISMRMRD_ACQUISITION)
    destination.write(message_id_bytes)
    acquisition.serialize_into(destination.write)


def write_waveform(destination, waveform):
    message_id_bytes = constants.GadgetMessageIdentifier.pack(constants.GADGET_MESSAGE_ISMRMRD_WAVEFORM)
    destination.write(message_id_bytes)
    waveform.serialize_into(destination.write)


def write_image(destination, image):
    message_id_bytes = constants.GadgetMessageIdentifier.pack(constants.GADGET_MESSAGE_ISMRMRD_IMAGE)
    destination.write(message_id_bytes)
    image.serialize_into(destination.write)





from . import util
from . import version
from . import external
from . import examples
from . import types

__version__ = version.version

# To maintain compatibility with the old Gadgetron interface, we need names/types to be available
# in the 'gadgetron' package. These are aliased here:
from .legacy import Gadget

from .types.image_array import ImageArray as IsmrmrdImageArray
from .types.recon_data import ReconData as IsmrmrdReconData
from .types.recon_data import ReconBit as IsmrmrdReconBit
from .types.recon_data import ReconBuffer as IsmrmrdDataBuffered
from .types.recon_data import SamplingDescription, SamplingLimit


__all__ = [
    util,
    legacy,
    external,
    examples,
    Gadget,
    IsmrmrdImageArray,
    IsmrmrdReconData,
    IsmrmrdReconBit,
    IsmrmrdDataBuffered,
    SamplingDescription,
    SamplingLimit
]




import os
import sys
import socket
import inspect
import logging

from typing import Callable

from .version import version
from .external import connection


def load_target(args) -> Callable[[connection.Connection], None]:

    module = __import__(args.get('module'), globals(), locals(), [args.get('target')], 0)
    target = getattr(module, args.get('target'), None)

    if inspect.isclass(target):
        return lambda conn: target().handle(conn)

    if callable(target):
        return target

    raise TypeError(f"Unable to initialize suitable target using symbol '{args.get('target')}'")


def main(args=None):

    if not args:
        return print(f"Gadgetron External Python Module v. {version}")

    args = dict(zip(['port', 'module', 'target'], args))

    logging.basicConfig(format=f"%(asctime)s.%(msecs)03d %(levelname)s " +
                               f"[ext. %(process)d {args.get('module')}.{args.get('target')}] %(message)s",
                        level=logging.getLevelName(os.environ.get('GADGETRON_EXTERNAL_LOG_LEVEL', 'DEBUG')),
                        datefmt="%m-%d %H:%M:%S")

    logging.debug(f"Starting external Python module '{args.get('module')}' in state: [ACTIVE]")
    logging.debug(f"Connecting to parent on port {args.get('port')}")

    address = ('localhost', int(args.get('port')))

    with connection.Connection(socket.create_connection(address, 30)) as conn:
        target = load_target(args)
        target(conn)


if __name__ == '__main__':
    try:
        main(sys.argv[1:])
    except Exception as e:
        logging.fatal(e)
        raise e




import socket
import logging

import xml.etree.ElementTree as xml

import ismrmrd

from . import constants

from .readers import read, read_byte_string, read_acquisition, read_waveform, read_image
from .writers import write_acquisition, write_waveform, write_image

from ..types.image_array import ImageArray, read_image_array, write_image_array
from ..types.recon_data import ReconData, read_recon_data, write_recon_data
from ..types.acquisition_bucket import read_acquisition_bucket


class Connection:
    """ Connection class representing a remote connection via the Gadgetron protocol

    """

    class SocketWrapper:

        def __init__(self, socket):
            self.socket = socket

        def read(self, nbytes):
            bytes = self.socket.recv(nbytes, socket.MSG_WAITALL)
            while len(bytes) < nbytes:
                bytes += self.read(nbytes - len(bytes))
            return bytes

        def write(self, byte_array):
            self.socket.sendall(byte_array)

        def close(self):
            end = constants.GadgetMessageIdentifier.pack(constants.GADGET_MESSAGE_CLOSE)
            self.socket.send(end)
            self.socket.close()

    class Struct:
        def __init__(self, **fields):
            self.__dict__.update(fields)

    def __init__(self, socket):
        self.socket = Connection.SocketWrapper(socket)

        self.readers = Connection._default_readers()
        self.writers = Connection._default_writers()

        self.raw_bytes = Connection.Struct(config=None, header=None)
        self.config, self.raw_bytes.config = self._read_config()
        self.header, self.raw_bytes.header = self._read_header()

        self.filters = []

    def __next__(self):
        return self.next()

    def __enter__(self):
        return self

    def __exit__(self, *exception_info):
        self.socket.close()

    def __iter__(self):
        while True:
            try:
                _, item = next(self)
                yield item
            except StopIteration:
                return

    def iter_with_mids(self):
        while True:
            try:
                yield next(self)
            except StopIteration:
                return

    def add_reader(self, slot, reader, *args, **kwargs):
        self.readers[slot] = lambda readable: reader(readable, *args, **kwargs)

    def add_writer(self, accepts, writer, *args, **kwargs):
        self.writers.insert(0, (accepts, lambda writable: writer(writable, *args, **kwargs)))

    def filter(self, predicate):
        """
         Filters the messages that come through the connection
        :param predicate: Function returning false if the message should be removed
        """
        if isinstance(predicate, type):
            return self.filters.append(lambda o: isinstance(o, predicate))
        self.filters.append(predicate)

    def send(self, item):
        """
        Sends a message via the connection
        :param item: Message to be sent. Must have corresponding writer
        """
        for predicate, writer in self.writers:
            if predicate(item):
                return writer(self.socket, item)
        raise TypeError(f"No appropriate writer found for item of type '{type(item)}'")

    def next(self):
        """
        :return: The next message from the connection
        """
        mid, item = self._read_item()

        while not all(pred(item) for pred in self.filters):
            self.send(item)
            mid, item = self._read_item()

        return mid, item

    def _read_item(self):
        message_identifier = self._read_message_identifier()

        def unknown_message_identifier(*_):
            logging.error(f"Received message (id: {message_identifier}) with no registered readers.")
            raise StopIteration()

        reader = self.readers.get(message_identifier, unknown_message_identifier)
        return message_identifier, reader(self.socket)

    def _read_message_identifier(self):
        return read(self.socket, constants.GadgetMessageIdentifier)

    def _read_config(self):
        message_identifier = self._read_message_identifier()
        assert(message_identifier == constants.GADGET_MESSAGE_CONFIG)
        config_bytes = read_byte_string(self.socket)
        return xml.fromstring(config_bytes), config_bytes

    def _read_header(self):
        message_identifier = self._read_message_identifier()
        assert(message_identifier == constants.GADGET_MESSAGE_HEADER)
        header_bytes = read_byte_string(self.socket)
        return ismrmrd.xsd.CreateFromDocument(header_bytes), header_bytes

    @ staticmethod
    def _default_readers():
        return {
            constants.GADGET_MESSAGE_CLOSE: Connection.stop_iteration,
            constants.GADGET_MESSAGE_ISMRMRD_ACQUISITION: read_acquisition,
            constants.GADGET_MESSAGE_ISMRMRD_WAVEFORM: read_waveform,
            constants.GADGET_MESSAGE_ISMRMRD_IMAGE: read_image,
            constants.GADGET_MESSAGE_IMAGE_ARRAY: read_image_array,
            constants.GADGET_MESSAGE_RECON_DATA: read_recon_data,
            constants.GADGET_MESSAGE_BUCKET: read_acquisition_bucket
        }

    @ staticmethod
    def _default_writers():
        return [
            (lambda item: isinstance(item, ismrmrd.Acquisition), write_acquisition),
            (lambda item: isinstance(item, ismrmrd.Waveform), write_waveform),
            (lambda item: isinstance(item, ismrmrd.Image), write_image),
            (lambda item: isinstance(item, ImageArray), write_image_array),
            (lambda item: isinstance(item, ReconData), write_recon_data)
        ]

    @ staticmethod
    def stop_iteration(_):
        logging.debug("Connection closed normally.")
        raise StopIteration()




import logging
import numpy as np
import gadgetron
import ismrmrd
import time
from multimethod import multimethod


def buffer_from_bucket(bucket, header):
    matrix_size = header.encoding[0].encodedSpace.matrixSize
    acqs = bucket.data

    logging.debug(f"Assembling buffer from bucket containing {len(acqs)} acquisitions.")

    number_of_channels = acqs[0].data.shape[0]
    number_of_samples = acqs[0].data.shape[1]
    buffer = np.zeros(
        (number_of_samples, matrix_size.y, matrix_size.z, number_of_channels, 1),
        dtype=np.complex64
    )

    for acq in acqs:
        buffer[:, acq.idx.kspace_encode_step_1, acq.idx.kspace_encode_step_2, :, 0] = acq.data.T

    return buffer, bucket.data[0].getHead()


def reconstruct_image(kspace_data):
    return gadgetron.util.cifftn(kspace_data, axes=[0, 1, 2])


def combine_channels(image):
    return np.sqrt(np.sum(np.abs(image) ** 2, axis=3))


def split_image_array(image_data):
    image_5d_view = image_data.reshape(*image_data.shape[:4], -1)
    return [image_5d_view[..., i] for i in range(image_5d_view.shape[4])]


def create_ismrmrd_image(image_data, reference_acquisition_header):
    return ismrmrd.image.Image.from_array(image_data, acquisition=reference_acquisition_header)


def reconstruct_buffer(kspace_data, reference_header):
    print(kspace_data.shape)
    image_data = reconstruct_image(kspace_data)
    image_data = combine_channels(image_data)
    image_data_list = split_image_array(image_data)
    images = [create_ismrmrd_image(img, reference_header) for img in image_data_list]
    return images

def flatten(lists):
    return sum(lists,[])

@multimethod
def reconstruct(buffers: gadgetron.types.ReconData, header):
    return flatten([[image for image in reconstruct_buffer(buffer.data.data, buffer.data.headers.flat[0])] for buffer in
            buffers])


@multimethod
def reconstruct(bucket: gadgetron.types.AcquisitionBucket, header):
    kspace_data, acq_header = buffer_from_bucket(bucket, header)
    return reconstruct_buffer(kspace_data, acq_header)


def send_images(connection, images):
    for img in images:
        print(img.getHead())
        connection.send(img)


def recon_buffers(connection):
    logging.info("Python reconstruction running - reconstructing images from acquisition buffers.")

    start = time.time()


    for data in connection:
        images = reconstruct(data, connection.header)
        send_images(connection, images)

    logging.info(f"Python reconstruction done. Duration: {(time.time() - start):.2f} s")


#######################################################################################################################
# This allows the file to be run directly as a server


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    gadgetron.external.listen(9100, recon_buffers)




import numpy
import struct
import ctypes

from gadgetron.external.readers import read, read_optional, read_array, read_object_array, read_acquisition_header
from gadgetron.external.writers import write_optional, write_array, write_object_array, write_acquisition_header
from gadgetron.external.constants import uint64, GadgetMessageIdentifier, GADGET_MESSAGE_RECON_DATA

uint16 = struct.Struct('<H')


class SamplingLimit(ctypes.Structure):
    _fields_ = [
        ('min', ctypes.c_uint16),
        ('center', ctypes.c_uint16),
        ('max', ctypes.c_uint16)
    ]

    def __iter__(self):
        return iter([self.min, self.center, self.max])


class SamplingDescription(ctypes.Structure):
    _fields_ = [
        ('encoded_FOV', ctypes.c_float * 3),
        ('recon_FOV', ctypes.c_float * 3),
        ('encoded_matrix', ctypes.c_uint16 * 3),
        ('recon_matrix', ctypes.c_uint16 * 3),
        ('sampling_limits', SamplingLimit * 3)
    ]


class ReconBuffer:
    def __init__(self, data, trajectory, density, headers, sampling_description):
        self.data = data
        self.trajectory = trajectory
        self.density = density
        self.headers = headers
        self.sampling = sampling_description


class ReconBit:
    def __init__(self, buffer, reference=None):
        self.data = buffer
        self.ref = reference


class ReconData:
    def __init__(self, bits):
        self.bits = bits

    def __getitem__(self, item):
        return self.bits[item]

    def __iter__(self):
        return iter(self.bits)


def read_sampling_description(source):
    return SamplingDescription.from_buffer_copy(source.read(ctypes.sizeof(SamplingDescription)))


def read_recon_buffer(source):

    data = read_array(source, numpy.complex64)
    trajectory = read_optional(source, read_array, numpy.float32)
    density = read_optional(source, read_array, numpy.float32)
    headers = read_object_array(source, read_acquisition_header)
    sampling_description = read_sampling_description(source)

    return ReconBuffer(data, trajectory, density, headers, sampling_description)


def read_recon_bit(source):
    buffer = read_recon_buffer(source)
    reference = read_optional(source, read_recon_buffer)
    return ReconBit(buffer, reference)


def read_recon_bits(source):
    size = read(source, uint64)
    return [read_recon_bit(source) for _ in range(size)]


def read_recon_data(source):
    return ReconData(read_recon_bits(source))


def write_sampling_description(destination, description):
    destination.write(description)


def write_recon_buffer(destination, buffer):
    write_array(destination, buffer.data)
    write_optional(destination, buffer.trajectory, write_array)
    write_optional(destination, buffer.density, write_array)
    write_object_array(destination, buffer.headers, write_acquisition_header)
    write_sampling_description(destination, buffer.sampling)


def write_recon_bit(destination, bit):
    write_recon_buffer(destination, bit.data)
    write_optional(destination, bit.ref, write_recon_buffer)


def write_recon_data(destination, recon_data):
    destination.write(GadgetMessageIdentifier.pack(GADGET_MESSAGE_RECON_DATA))
    destination.write(uint64.pack(len(recon_data.bits)))
    for bit in recon_data.bits:
        write_recon_bit(destination, bit)




try:
    import pyfftw.interfaces.numpy_fft as fft
except ImportError:
    import numpy.fft as fft


def cfftn(data, axes):
    """
    Centered fast fourier transform, n-dimensional

    :param data: complex input data
    :param axes: Axes along which to shift
    :return: fourier transformed data
    """
    return fft.fftshift(fft.fftn(fft.ifftshift(data, axes=axes), axes=axes, norm='ortho'), axes=axes)


def cifftn(data, axes):
    """
    Centered inverse fast fourier transform, n-dimensional

    :param data: complex input data
    :param axes: Axes along which to shift
    :return: inverse fourier transformed data
    """
    return fft.ifftshift(fft.ifftn(fft.fftshift(data, axes=axes), axes=axes, norm='ortho'), axes=axes)



import struct

GADGET_MESSAGE_INT_ID_MIN                             =    0
GADGET_MESSAGE_FILENAME                               =    1
GADGET_MESSAGE_CONFIG                                 =    2
GADGET_MESSAGE_HEADER                                 =    3
GADGET_MESSAGE_CLOSE                                  =    4
GADGET_MESSAGE_TEXT                                   =    5
GADGET_MESSAGE_QUERY                                  =    6
GADGET_MESSAGE_RESPONSE                               =    7
GADGET_MESSAGE_ERROR                                  =    8
GADGET_MESSAGE_INT_ID_MAX                             =  999
GADGET_MESSAGE_EXT_ID_MIN                             = 1000
GADGET_MESSAGE_ISMRMRD_ACQUISITION                    = 1008
GADGET_MESSAGE_ISMRMRD_IMAGE_CPLX_FLOAT               = 1009
GADGET_MESSAGE_ISMRMRD_IMAGE_REAL_FLOAT               = 1010
GADGET_MESSAGE_ISMRMRD_IMAGE_REAL_USHORT              = 1011
GADGET_MESSAGE_DICOM                                  = 1012
GADGET_MESSAGE_ISMRMRD_IMAGE                          = 1022
GADGET_MESSAGE_RECON_DATA                             = 1023
GADGET_MESSAGE_IMAGE_ARRAY                            = 1024
GADGET_MESSAGE_ISMRMRD_WAVEFORM                       = 1026
GADGET_MESSAGE_BUCKET                                 = 1050
GADGET_MESSAGE_EXT_ID_MAX                             = 4096

GadgetMessageIdentifier = struct.Struct('<H')

bool = struct.Struct('<?')
uint32 = struct.Struct('<I')
uint64 = struct.Struct('<Q')
uint16 = struct.Struct('<H')



from setuptools import setup

import gadgetron

setup(
    name='gadgetron',
    version=gadgetron.__version__,
    packages=[
        'gadgetron',
        'gadgetron.util',
        'gadgetron.external',
        'gadgetron.examples',
        'gadgetron.legacy',
        'gadgetron.types'
    ],
    url='',
    license='MIT',
    author='Kristoffer Langeland Knudsen',
    author_email='kristofferlknudsen@gradientsoftware.net',
    description='',
    install_requires=[
        'numpy>=1.15.1',
        'ismrmrd>=1.6.2',
        'pyFFTW>=0.11',
        'multimethod'
    ]
)




import time
import logging
import ismrmrd

import numpy as np

from gadgetron.util.cfft import cfftn, cifftn


def noise_adjustment(acquisitions, header):
    # The dataset might include noise measurements (mine does). We'll consume noise measurements, use them
    # to prepare a noise adjustment matrix, and never pass them down the chain. They contain no image data,
    # and will not be missed. We'll also perform noise adjustment on following acquisitions, when we have a
    # noise matrix available.

    noise_matrix = None
    noise_dwell_time = 1.0

    try:
        noise_bandwidth = header.encoding[0].acquisitionSystemInformation.relativeNoiseBandwidth
    except:
        noise_bandwidth = 0.793

    def scaling_factor(acq):
        return np.sqrt(2 * acq.sample_time_us * noise_bandwidth / noise_dwell_time)

    def calculate_whitening_transformation(noise):
        noise = np.asmatrix(noise)
        covariance = (1.0 / (noise.shape[1] - 1)) * (noise * noise.H)
        return np.linalg.inv(np.linalg.cholesky(covariance))

    def apply_whitening_transformation(acq):
        return np.asarray(scaling_factor(acq) * noise_matrix * np.asmatrix(acq.data))

    def noise_adjust(acq):
        if noise_matrix is not None:
            acq.data[:] = apply_whitening_transformation(acq)
        return acq

    for acquisition in acquisitions:
        if acquisition.is_flag_set(ismrmrd.ACQ_IS_NOISE_MEASUREMENT):
            noise_matrix = calculate_whitening_transformation(acquisition.data)
            noise_dwell_time = acquisition.sample_time_us
        else:
            yield noise_adjust(acquisition)


def remove_oversampling(acquisitions, header):
    # The dataset I'm working with was originally taken on a Siemens scanner. It features 2x oversampling
    # along the first image dimension. We're going to have a look at the header. If our encoded space
    # doesn't match the recon space, we're going to crop the acquisition.

    encoding_space = header.encoding[0].encodedSpace.matrixSize
    recon_space = header.encoding[0].reconSpace.matrixSize

    if encoding_space.x == recon_space.x:
        return acquisitions

    x0 = (encoding_space.x - recon_space.x) // 2
    x1 = (encoding_space.x - recon_space.x) // 2 + recon_space.x

    def crop_acquisition(acquisition):
        x_space = cifftn(acquisition.data, axes=[1])
        x_space = x_space[:, x0:x1]
        acquisition.resize(number_of_samples=x_space.shape[1], active_channels=x_space.shape[0])
        acquisition.center_sample = recon_space.x // 2
        acquisition.data[:] = cfftn(x_space, axes=[1])

        return acquisition

    return map(crop_acquisition, acquisitions)


def accumulate_acquisitions(acquisitions, header):
    # To form images, we need a full slice of data. We accumulate acquisitions until we reach the
    # end of a slice. The acquisitions are then combined in a single buffer, which is passed
    # on. We also pass on a reference acquisition. We use it later to initialize image metadata.

    accumulated_acquisitions = []
    matrix_size = header.encoding[0].encodedSpace.matrixSize

    def assemble_buffer(acqs):
        logging.debug(f"Assembling buffer from {len(acqs)} acquisitions.")

        number_of_channels = acqs[0].data.shape[0]
        number_of_samples = acqs[0].data.shape[1]

        buffer = np.zeros(
            (number_of_channels,
             number_of_samples,
             matrix_size.y,
             matrix_size.z),
            dtype=np.complex64
        )

        for acq in acqs:
            buffer[:, :, acq.idx.kspace_encode_step_1, acq.idx.kspace_encode_step_2] = acq.data

        return buffer

    for acquisition in acquisitions:
        accumulated_acquisitions.append(acquisition)
        if acquisition.is_flag_set(ismrmrd.ACQ_LAST_IN_SLICE):
            yield acquisition, assemble_buffer(accumulated_acquisitions)
            accumulated_acquisitions = []


def reconstruct_images(buffers):

    def reconstruct_image(kspace_data):
        # Reconstruction is an inverse fft in this case.

        return cifftn(kspace_data, axes=[1, 2, 3])

    def combine_channels(image_data):
        # The buffer contains complex images, one for each channel. We combine these into a single image
        # through a sum of squares along the channels (axis 0).

        return np.sqrt(np.sum(np.square(np.abs(image_data)), axis=0))

    for reference, data in buffers:
        yield ismrmrd.image.Image.from_array(
            combine_channels(reconstruct_image(data)),
            acquisition=reference
        )


def recon_acquisitions(connection):
    logging.info("Python reconstruction running - reconstructing images from acquisitions.")

    start = time.time()

    # We're only interested in acquisitions in this example, so we filter the connection. Anything filtered out in
    # this way will pass back to Gadgetron unchanged.
    connection.filter(ismrmrd.Acquisition)

    # Connections are iterable - iterating them can be done once, and will yield the data sent from Gadgetron.
    acquisitions = iter(connection)
    acquisitions = noise_adjustment(acquisitions, connection.header)
    acquisitions = remove_oversampling(acquisitions, connection.header)
    buffers = accumulate_acquisitions(acquisitions, connection.header)
    images = reconstruct_images(buffers)

    for image in images:
        logging.debug("Sending image back to client.")
        connection.send(image)

    logging.info(f"Python reconstruction done. Duration: {(time.time() - start):.2f} s")




import numpy
import ismrmrd
import logging

from ..external import constants
from ..external import handlers


def _transform_to_legacy_acquisition(acquisition):
    return acquisition.getHead(), numpy.transpose(acquisition.data)


def _transform_from_legacy_acquisition(items):
    header, data = items
    header.number_of_samples, header.active_channels = data.shape
    acquisition = ismrmrd.Acquisition(header)
    acquisition.data[:] = numpy.transpose(data)
    return acquisition


def _transform_to_legacy_waveform(waveform):
    return waveform,


def _transform_to_legacy_image(image):
    return image.getHead(), numpy.transpose(image.data), image.attribute_string


def _transform_from_legacy_image(items):
    items = list(items) + [""]
    header, data, meta = items[:3]
    header.attribute_string_len = len(meta)
    image = ismrmrd.Image(header, meta)
    image.data[:] = numpy.reshape(numpy.transpose(data), image.data.shape)
    return image


def _parse_params(xml):
    return {p.get('name'): p.get('value') for p in xml.iter('property')}


class Gadget:

    _reader_transformations = {
        constants.GADGET_MESSAGE_ISMRMRD_ACQUISITION: _transform_to_legacy_acquisition,
        constants.GADGET_MESSAGE_ISMRMRD_IMAGE: _transform_to_legacy_image
    }

    _writer_transformation = [
        (lambda args: isinstance(args[0], ismrmrd.AcquisitionHeader), _transform_from_legacy_acquisition),
        (lambda args: isinstance(args[0], ismrmrd.ImageHeader), _transform_from_legacy_image)
    ]

    @ classmethod
    def __init_subclass__(cls, **kwargs):
        handlers.register_handler(lambda conn: cls().handle(conn))

    def __init__(self, *args, **kwargs):
        self.connection = None
        self.params = None
        self.hooks = {
            constants.GADGET_MESSAGE_ISMRMRD_WAVEFORM:
                self._process_waveform if hasattr(self, 'process_waveform') else self._ignore_waveform
        }

    def handle(self, connection):
        self.connection = connection
        self.params = _parse_params(connection.config)
        self.process_config(connection.raw.header)

        def invoke_process(process, args):
            if not args:
                raise ValueError()
            try:
                process(*args)
            except TypeError:
                invoke_process(process, args[:-1])

        for mid, item in connection.iter_with_mids():
            if mid in self.hooks:
                self.hooks.get(mid)(item)
            else:
                transformation = Gadget._reader_transformations.get(mid, lambda x: [x])
                args = transformation(item)
                try:
                    invoke_process(self.process, args)
                except ValueError:
                    raise TypeError(f"Failed to invoke self.process; arguments do not match. Had arguments: "
                                    f"{[type(a) for a in args]}")

    def process_config(self, config):
        pass

    def process(self, *args):
        pass

    def wait(self):
        pass

    def put_next(self, *args):
        transformation = next((trans for pred, trans in Gadget._writer_transformation if pred(args)), lambda x: x[0])
        self.connection.send(transformation(args))

    def _ignore_waveform(self, waveform):
        self.connection.send(waveform)

    def _process_waveform(self, waveform):
        self.process_waveform(waveform)




import numpy

from ..external import readers
from ..external import writers
from ..external import constants


class ImageArray:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def read_meta_container(source):
    return readers.read_byte_string(source, constants.uint64).decode('utf-8')


def read_meta_container_vector(source):
    size = readers.read(source, constants.uint64)
    return [read_meta_container(source) for _ in range(size)]


def read_waveforms(source):
    size = readers.read(source, constants.uint64)
    return [readers.read_waveform(source) for _ in range(size)]


def read_image_array(source):
    return ImageArray(
        data=readers.read_array(source, numpy.complex64),
        headers=readers.read_object_array(source, readers.read_image_header),
        meta=read_meta_container_vector(source),
        waveform=readers.read_optional(source, read_waveforms),
        acq_headers=readers.read_optional(source, readers.read_object_array, readers.read_acquisition_header)
    )


def write_meta_container(destination, container):
    writers.write_byte_string(destination, container.encode('utf-8'), constants.uint64)


def write_meta_container_vector(destination, containers):
    destination.write(constants.uint64.pack(len(containers)))
    for container in containers:
        write_meta_container(destination, container)


def write_waveforms(destination, waveforms):
    destination.write(constants.uint64.pack(len(waveforms)))
    for waveform in waveforms:
        writers.write_waveform(destination, waveform)


def write_image_array(destination, image_array):
    destination.write(constants.GadgetMessageIdentifier.pack(constants.GADGET_MESSAGE_IMAGE_ARRAY))
    writers.write_array(destination, image_array.data)
    writers.write_object_array(destination, image_array.header, writers.write_image_header)
    write_meta_container_vector(destination, image_array.meta)
    writers.write_optional(destination, image_array.waveform, write_waveforms)
    writers.write_optional(destination, image_array.acq_headers, writers.write_object_array, writers.write_acquisition_header)





import ctypes
import ismrmrd

import functools

import numpy

from . import constants


def read(source, type):
    return type.unpack(source.read(type.size))[0]


def read_optional(source, continuation, *args, **kwargs):
    is_present = read(source, constants.bool)
    return continuation(source, *args, **kwargs) if is_present else None


def read_vector(source, numpy_type=numpy.uint64):
    size = read(source, constants.uint64)
    dtype = numpy.dtype(numpy_type)
    return numpy.frombuffer(source.read(size * dtype.itemsize), dtype=dtype)


def read_array(source, numpy_type=numpy.uint64):
    dtype = numpy.dtype(numpy_type)
    dimensions = read_vector(source)
    elements = int(functools.reduce(lambda a, b: a * b, dimensions))
    return numpy.reshape(numpy.frombuffer(source.read(elements * dtype.itemsize), dtype=dtype), dimensions, order='F')


def read_object_array(source, read_object):
    dimensions = read_vector(source)
    elements = int(functools.reduce(lambda a, b: a * b, dimensions))
    return numpy.reshape(numpy.asarray([read_object(source) for _ in range(elements)], dtype=object), dimensions)


def read_image_header(source):
    header_bytes = source.read(ctypes.sizeof(ismrmrd.ImageHeader))
    return ismrmrd.ImageHeader.from_buffer_copy(header_bytes)


def read_acquisition_header(source):
    header_bytes = source.read(ctypes.sizeof(ismrmrd.AcquisitionHeader))
    return ismrmrd.AcquisitionHeader.from_buffer_copy(header_bytes)


def read_waveform_header(source):
    header_bytes = source.read(ctypes.sizeof(ismrmrd.WaveformHeader))
    return ismrmrd.Waveform.from_buffer_copy(header_bytes)


def read_gadget_message_length(source, type=constants.uint32):
    return read(source, type)


def read_byte_string(source, type=constants.uint32):
    length = read_gadget_message_length(source, type)
    byte_string = source.read(length)
    return byte_string


def read_acquisition(source):
    return ismrmrd.Acquisition.deserialize_from(source.read)


def read_waveform(source):
    return ismrmrd.Waveform.deserialize_from(source.read)


def read_image(source):
    return ismrmrd.Image.deserialize_from(source.read)
